# declares public API for this module
# loads functions from subscripts which are needed in main.py
import refinegems.databases
import refinegems.biomass
import refinegems.growth
import refinegems.modelseed
import refinegems.sboann
import refinegems.charges
import refinegems.pathways
import refinegems.curate
import refinegems.polish
import refinegems.cvterms
import refinegems.analysis_db
import refinegems.analysis_biocyc
import refinegems.analysis_kegg
import refinegems.entities
import refinegems.gapfill
import refinegems.comparison
import refinegems.investigate
import refinegems.io

__author__ = "Famke Baeuerle"



#!/usr/bin/env python
""" Provides functions to compare genes found in KEGG and in the model

Extracts all KEGG ids from the annotations and compares them to a list for you organism from KEGG.
Reactions with KEGG ids not found in the model are expanded to a table containing the KEGG id,
the locus_tag (old format), the EC number, the BiGG id and the Gene Protein Reaction (GPR) rule.
This section needs a gff file of your organism, the txt file from BiGG containing all reactions
and the KEGG identifier of your organism.

Due to the KEGG REST API this is relatively slow (model of size 1500 reactions - 20 min).
"""

import pandas as pd
from bioservices.kegg import KEGG
import io
import re
from libsbml import Model as libModel
from refinegems.io import parse_gff_for_gp_info
from refinegems.entities import get_model_genes, compare_gene_lists, get_model_reacs_or_metabs
from refinegems.analysis_db import get_bigg2other_db, compare_bigg_model

__author__ = "Famke Baeuerle"


def get_kegg_genes(organismid: str) -> pd.DataFrame:
    """Extracts list of genes from KEGG given an organism

    Args:
        - organismid (str): KEGG ID of organism which the model is based on

    Returns:
        pd.DataFrame: Table of all genes denoted in KEGG for the organism
    """
    k = KEGG()
    gene_list = k.list(organismid)

    return pd.read_table(io.StringIO(gene_list), header=None)


def get_locus_ec(genes_kegg_notmodel: pd.DataFrame) -> pd.DataFrame:
    """Creates columns with EC numbers for the locus tags of the genes

    Args:
        - genes_kegg_notmodel (pd.DataFrame): Genes present in KEGG but not in the model

    Returns:
        pd.DataFrame: Table of genes with locus tag and EC number
    """
    k = KEGG()

    ec_dict = {}
    for gene in genes_kegg_notmodel:
        entry = k.parse(k.get(gene))
        try:
            ec_dict[entry['ENTRY']] = (entry['ORTHOLOGY'])
        except(KeyError):
            pass

    real_ec = {}
    for entry, ortho in ec_dict.items():
        for key, value in ortho.items():
            m = re.search('(?:EC).*', value)
            if m:
                real_ec[entry[:12]] = '[' + m.group(0)

    locus_ec = pd.DataFrame.from_dict(
        real_ec,
        orient='index').reset_index().rename(
        columns={
            'index': 'locus_tag',
            0: 'EC-number'})

    def slice_ec(ec):
        new = ec[4:]
        new2 = new[:-1]
        return new2

    locus_ec['EC'] = locus_ec.apply(
        lambda row: slice_ec(
            row['EC-number']), axis=1)
    locus_ec = locus_ec.drop('EC-number', axis=1)

    return locus_ec


def get_locus_ec_kegg(locus_ec: pd.DataFrame) -> pd.DataFrame:
    """Searches for KEGG reactions based on EC numbers

    Args:
        - locus_ec (pd.DataFrame): Genes with locus tag and EC number

    Returns:
        pd.DataFrame: Table of genes with locus tag, EC number and KEGG Id
    """

    def get_kegg_reaction(ec_number):
        k = KEGG()
        gene = k.parse(k.get(ec_number))
        try:
            return gene['REACTION'][-1]
        except(KeyError):
            pass
        return None

    def drop_nonreac(kegg_id):
        if len(kegg_id) != 11:
            return None
        else:
            return kegg_id

    def slice_kegg(kegg):
        return kegg[4:-1]

    locus_ec['KEGG_Ids'] = locus_ec.apply(
        lambda row: get_kegg_reaction(
            row['EC']), axis=1)
    locus_ec = locus_ec.dropna()
    locus_ec.loc[:, 'KEGG_Ids2'] = locus_ec.apply(
        lambda row: drop_nonreac(
            row['KEGG_Ids']), axis=1)
    locus_ec = locus_ec.dropna()
    locus_ec['KEGG'] = locus_ec.apply(
        lambda row: slice_kegg(
            row['KEGG_Ids2']), axis=1)
    locus_ec_kegg = locus_ec.dropna().drop(
        'KEGG_Ids', axis=1).drop(
        'KEGG_Ids2', axis=1)

    return locus_ec_kegg


def get_locus_ec_kegg_bigg(locus_ec_kegg: pd.DataFrame, bigg_kegg: pd.DataFrame) -> pd.DataFrame:
    """Merges table with genes from model with BiGG / KEGG mapping to add BiGG Ids

    Args:
        - locus_ec_kegg (pd.DataFrame): Genes with locus tag, EC number and KEGG Id
        - bigg_kegg (pd.DataFrame): BiGG IDs with corresponding KEGG Ids

    Returns:
        pd.DataFrame: Table of genes with locus tag, EC number, KEGG Id and BiGG Id
    """
    locus_ec_kegg_bigg = locus_ec_kegg.merge(bigg_kegg, on=['KEGG'])
    return locus_ec_kegg_bigg


def get_locus_ec_kegg_bigg_gpr(locus_ec_kegg_bigg: pd.DataFrame, locus_gpr: pd.DataFrame) -> pd.DataFrame:
    """Merges table with genes from model if locus tag / GPR mapping to add GPRs

    Args:
        - locus_ec_kegg_bigg (pd.DataFrame): Genes with locus tag, EC number, KEGG Id and BiGG Id
        - locus_gpr (pd.DataFrame): Mapping from locus tags to GPRs

    Returns:
        pd.DataFrame: Table of genes with locus tag, EC number, KEGG Id, BiGG Id and GPR
    """

    def slice_locus(locus):
        return locus[:-1]

    locus_ec_kegg_bigg['locus_tag'] = locus_ec_kegg_bigg.apply(
        lambda row: slice_locus(row['locus_tag']), axis=1)

    return locus_ec_kegg_bigg.merge(locus_gpr, how='left', on='locus_tag')


def kegg_gene_comp(model: libModel, organismid: str, gff_file: str) -> pd.DataFrame:
    """Exectues all steps to compare genes of the model to KEGG genes

    Args:
        - model (libModel): Model loaded with libSBML
        - organismid (str): KEGG ID of organism which the model is based on
        - gff_file (str): Path to gff file of organism of interest

    Returns:
        pd.DataFrame: Table containing missing reactions with locus tag, EC number, KEGG Id, BiGG Id and GPR
    """
    model_genes = get_model_genes(model, True)
    model_reactions = get_model_reacs_or_metabs(model)
    kegg_genes = get_kegg_genes(organismid)
    bigg_kegg = get_bigg2other_db('KEGG')
    genes_kegg_notmodel = compare_gene_lists(model_genes, kegg_genes)
    locus_gpr = parse_gff_for_gp_info(gff_file)
    locus_ec = get_locus_ec(genes_kegg_notmodel)
    locus_ec_kegg = get_locus_ec_kegg(locus_ec)
    locus_ec_kegg_bigg = get_locus_ec_kegg_bigg(locus_ec_kegg, bigg_kegg)
    locus_ec_kegg_bigg_gpr = get_locus_ec_kegg_bigg_gpr(
        locus_ec_kegg_bigg, locus_gpr)
    missing_reactions = compare_bigg_model(
        locus_ec_kegg_bigg_gpr, model_reactions)
    return missing_reactions



#!/usr/bin/env python
""" Provides functions to simulate growth on any medium

Tailored to work with the media denoted in the local db, should work with any medium as long as its defined in a csv with ; as delimiter and BiGG Ids for the compounds. Use refinegems.io.load_medium_custom and hand this to the growth_one_medium_from_default or growth_one_medium_from_minimum function.
"""

import logging
import pandas as pd
import numpy as np
from refinegems.io import load_medium_from_db
from cobra.medium import minimal_medium
from cobra import Reaction
from cobra import Model as cobraModel

__author__ = "Famke Baeuerle"


def set_fluxes_to_simulate(reaction: Reaction) -> Reaction:
    """Helper function: Set flux bounds to -1000.0 and 1000.0 to enable model simulation with growth_one_medium_from_minimal/default

    Args:
        - reaction (Reaction): Reaction with unusable flux bounds

    Returns:
        Reaction: Reaction with usable flux bounds
    """
    reaction.bounds = (-1000.0, 1000.0)
    return reaction


def get_default_uptake(model: cobraModel) -> list[str]:
    """Determines which metabolites are used in the standard medium

    Args:
        - model (cobraModel): Model loaded with COBRApy

    Returns:
        list[str]: Metabolites consumed in standard medium
    """
    with model:
        sol = model.optimize()
        fluxes = sol.fluxes
        default_uptake = []
        for index, value in fluxes.items():
            if "EX" in index:
                if value < 0:
                    default_uptake.append(index)
    return default_uptake


def get_minimal_uptake(model: cobraModel) -> list[str]:
    """Determines which metabolites are used in a minimal medium

    Args:
        - model (cobraModel): Model loaded with COBRApy

    Returns:
        list[str]: Metabolites consumed in minimal medium
    """
    with model:
        minimal = minimal_medium(model)
    return list(minimal.index)


def get_default_secretion(model: cobraModel) -> list[str]:
    """Checks fluxes after FBA, if positive the metabolite is produced

    Args:
        - model (cobraModel): Model loaded with COBRApy

    Returns:
        list[str]: BiGG Ids of produced metabolites
    """
    with model:
        fluxes = model.optimize().fluxes
        default_secretion = []
        for index, value in fluxes.items():
            if value > 0:
                default_secretion.append(index)
    return default_secretion


def get_missing_exchanges(model: cobraModel, medium: pd.DataFrame) -> list[str]:
    """Look for exchange reactions needed by the medium but not in the model

    Args:
        - model (cobraModel): Model loaded with COBRApy
        - medium (pd.DataFrame): Dataframe with medium definition

    Returns:
        list[str]: Ids of all exchanges missing in the model but given in medium
    """
    medium_list = medium['BiGG_EX'].dropna().tolist()
    missing_exchanges = []
    for exchange in medium_list:
        try:
            model.reactions.get_by_id(exchange)
        except(KeyError):
            missing_exchanges.append(exchange)
    return missing_exchanges


def modify_medium(medium: pd.DataFrame, missing_exchanges: list[str]) -> dict:
    """Helper function: Remove exchanges from medium that are not in the model to avoid KeyError

    Args:
        - medium (pd.DataFrame): Dataframe with medium definition
        - missing_exchanges (list): Ids of exchanges not in the model

    Returns:
        dict: Growth medium definition that can be used with the model (f.ex {'EX_glc__D_e' : 10.0})
    """
    growth_medium = dict.fromkeys(medium['BiGG_EX'].dropna().tolist(), 10.0)
    for exchange in missing_exchanges:
        growth_medium.pop(exchange)
    return growth_medium


def find_missing_essential(model: cobraModel, growth_medium: dict, default_uptake: list[str], anaerobic: bool) -> list[str]:
    """Report which exchange reactions are needed for growth, combines default uptake and valid new medium

    Args:
        - model (cobraModel): Model loaded with COBRApy
        - growth_medium (dict): Growth medium definition that can be used with the model. Output of modify_medium.
        - default_uptake (list[str]): Metabolites consumed in standard medium
        - anaerobic (bool): If True 'EX_o2_e' is set to 0.0 to simulate anaerobic conditions

    Returns:
        list[str]: Ids of exchanges of all metabolites which lead to zero growth if blocked
    """
    with model:
        default_medium = {i: 10.0 for i in default_uptake}
        if anaerobic and ('EX_o2_e' in default_medium): default_medium['EX_o2_e'] = 0.0
        if anaerobic and ('EX_o2_e' in growth_medium): growth_medium['EX_o2_e'] = 0.0
        new_medium = {**growth_medium, **default_medium}
        try:
            model.medium = new_medium
        except(ValueError):
            logging.info('Change upper bounds to 1000.0 and lower bounds to -1000.0 to make model simulatable.')
            for reaction in model.reactions:
                set_fluxes_to_simulate(reaction)
            model.medium = new_medium
        essential = []
        for metab in new_medium.keys():
            with model:
                model.reactions.get_by_id(metab).lower_bound = 0
                sol = model.optimize()
                if sol.objective_value < 1e-5:  # and sol.objective_value > -1e-9: # == 0 no negative growth!
                    essential.append(metab)
                model.reactions.get_by_id(metab).lower_bound = -10
    return essential


def find_minimum_essential(medium: pd.DataFrame, essential: list[str]) -> list[str]:
    """Report metabolites necessary for growth and not in custom medium

    Args:
        - medium (pd.DataFrame): Dataframe with medium definition
        - essential (list[str]): Ids of all metabolites which lead to zero growth if blocked. Output of find_missing_essential.

    Returns:
        list[str]: Ids of exchanges of metabolites not present in the medium but necessary for growth
    """
    minimum = []
    for metab in essential:
        if metab not in medium['BiGG_EX'].tolist():
            minimum.append(metab)
    return minimum


def simulate_minimum_essential(model: cobraModel, growth_medium: dict, minimum: list[str], anaerobic: bool) -> float:
    """Simulate growth with custom medium plus necessary uptakes

    Args:
        - model (cobraModel): Model loaded with COBRApy
        - growth_medium (dict): Growth medium definition that can be used with the model. Output of modify_medium.
        - minimum (list[str]): Ids of exchanges of metabolites not present in the medium but necessary for growth. Output of find_minimum_essential.
        - anaerobic (bool): If True 'EX_o2_e' is set to 0.0 to simulate anaerobic conditions

    Returns:
        float: Growth value in mmol per (gram dry weight) per hour
    """
    with model:
        min_medium = {i: 10.0 for i in minimum}
        new_medium = {**growth_medium, **min_medium}
        try:
            if (new_medium['EX_o2_e'] == 10.0):
                new_medium['EX_o2_e'] = 20.0 if not anaerobic else 0.0
        except KeyError:
            print('No Oxygen Exchange Reaction')
            pass
        try:
            model.medium = new_medium
        except(ValueError):
            logging.info('Change upper bounds to 1000.0 and lower bounds to -1000.0 to make model simulatable.')
            for reaction in model.reactions:
                set_fluxes_to_simulate(reaction)
            model.medium = new_medium
        sol = model.optimize()
    return sol.objective_value


def get_all_minimum_essential(model: cobraModel, media: list[str]) -> pd.DataFrame:
    """Returns metabolites necessary for growth and not in media

    Args:
        - model (cobraModel): Model loaded with COBRApy
        - media (list[str]): Containing the names of all media for which the growth essential metabolites not contained in the media should be returned

    Returns:
        pd.DataFrame: information on different media which metabs are missing
    """
    default_uptake = get_default_uptake(model)
    mins = pd.DataFrame()
    for medium in media:
        medium_df = load_medium_from_db(medium)
        missing_exchanges = get_missing_exchanges(model, medium_df)
        medium_dict = modify_medium(medium_df, missing_exchanges)
        essential = find_missing_essential(model, medium_dict, default_uptake)
        minimum = find_minimum_essential(medium_df, essential)
        mins[medium['medium'][0]] = pd.Series(minimum)
    return mins


def growth_one_medium_from_default(model: cobraModel, medium: pd.DataFrame, anaerobic: bool) -> pd.DataFrame:
    """Simulates growth on given medium, adding missing metabolites from the default uptake

    Args:
        - model (cobraModel): Model loaded with COBRApy
        - medium (pd.DataFrame): Dataframe with medium definition
        - anaerobic (bool): If True 'EX_o2_e' is set to 0.0 to simulate anaerobic conditions

    Returns:
        pd.DataFrame: Information on growth behaviour on given medium
    """
    default_uptake = get_default_uptake(model)
    missing_exchanges = get_missing_exchanges(model, medium)
    medium_dict = modify_medium(medium, missing_exchanges)
    essential = find_missing_essential(model, medium_dict, default_uptake, anaerobic)
    minimum = find_minimum_essential(medium, essential)

    medium_dict = modify_medium(medium, missing_exchanges)
    growth_value = simulate_minimum_essential(model, medium_dict, minimum, anaerobic)
    doubling_time = (np.log(2) / growth_value) * 60
    medium_name = medium['medium'][0] if not anaerobic else f'{medium["medium"][0]}[-O2]'
    exchanges = [[medium_name], minimum,
                 missing_exchanges, [growth_value], [doubling_time]]
    df_growth = pd.DataFrame(exchanges,
                             ['medium',
                              'essential',
                              'missing exchanges',
                              'growth_value',
                              'doubling_time [min]']).T
    return df_growth


def growth_one_medium_from_minimal(model: cobraModel, medium: pd.DataFrame, anaerobic: bool) -> pd.DataFrame:
    """Simulates growth on given medium, adding missing metabolites from a minimal uptake

    Args:
        - model (cobraModel): Model loaded with COBRApy
        - medium (pd.DataFrame): Dataframe with medium definition
        - anaerobic (bool): If True 'EX_o2_e' is set to 0.0 to simulate anaerobic conditions

    Returns:
        pd.DataFrame: Information on growth behaviour on given medium
    """
    minimal_uptake = get_minimal_uptake(
        model)  # use this instead of default_uptake
    missing_exchanges = get_missing_exchanges(model, medium)
    medium_dict = modify_medium(medium, missing_exchanges)
    essential = find_missing_essential(model, medium_dict, minimal_uptake, anaerobic)
    minimum = find_minimum_essential(medium, essential)

    medium_dict = modify_medium(medium, missing_exchanges)
    growth_value = simulate_minimum_essential(model, medium_dict, minimum, anaerobic)
    doubling_time = (np.log(2) / growth_value) * 60
    medium_name = medium['medium'][0] if not anaerobic else f'{medium["medium"][0]}[-O2]'
    exchanges = [[medium_name], minimum,
                 missing_exchanges, [growth_value], [doubling_time]]
    df_growth = pd.DataFrame(exchanges,
                             ['medium',
                              'essential',
                              'missing exchanges',
                              'growth_value',
                              'doubling_time [min]']).T
    return df_growth


def get_growth_selected_media(model: cobraModel, media: list[str], basis: str, anaerobic: bool) -> pd.DataFrame:
    """Simulates growth on all given media

    Args:
        - model (cobraModel): Model loaded with COBRApy
        - media (list[str]): Ids of media to simulate on
        - basis (str): Either default_uptake (adding metabs from default) or minimal_uptake (adding metabs from minimal medium)
        - anaerobic (bool): If True 'EX_o2_e' is set to 0.0 to simulate anaerobic conditions

    Returns:
        pd.DataFrame: Information on growth behaviour on given media
    """
    growth = pd.DataFrame()
    for medium in media:
        medium_df = load_medium_from_db(medium)
        if (basis == 'default_uptake'):
            growth_one = growth_one_medium_from_default(model, medium_df, anaerobic)
        elif (basis == 'minimal_uptake'):
            growth_one = growth_one_medium_from_minimal(model, medium_df, anaerobic)
        growth = growth.append(growth_one, ignore_index=True)
    return growth


def get_essential_reactions(model: cobraModel) -> list[str]:
    """Knocks out each reaction, if no growth is detected the reaction is seen as essential

    Args:
        - model (cobraModel): Model loaded with COBRApy

    Returns:
        list[str]: BiGG Ids of essential reactions
    """
    ess = []
    for reaction in model.reactions:
        with model as model:
            reaction.knock_out()
            model.optimize()
            if model.objective.value <= 11:
                print('%s blocked (bounds: %s), new growth rate %f $' %
                      (reaction.id, str(reaction.bounds), model.objective.value))
                ess.append(reaction.id)

    return ess


def get_essential_reactions_via_bounds(model: cobraModel) -> list[str]:
    """Knocks out reactions by setting their bounds to 0, if no growth is detected the reaction is seen as essential


    Args:
        - model (cobraModel): Model loaded with COBRApy

    Returns:
        list[str]: BiGG Ids of essential reactions
    """
    medium = model.medium
    ess = []
    for content in medium.keys():
        model.reactions.get_by_id(content).lower_bound = 0.0
        solution = model.optimize().objective_value
        if solution < 1e-9:
            ess.append(content)
        model.reactions.get_by_id(content).lower_bound = -10.0

    return ess


def find_additives(model:cobraModel, base_medium: dict) -> pd.DataFrame:
    """Iterates through all exchanges to find metabolites that lead to a higher growth rate compared to the growth rate yielded on the base_medium

    Args:
        - model (cobraModel): Model loaded with COBRApy
        - base_medium (dict): Exchanges as keys and their flux bound as value (f.ex {'EX_glc__D_e' : 10.0})

    Returns:
        pd.DataFrame: Exchanges sorted from highest to lowest growth rate improvement
    """
    with model:
        medium = model.medium
        model.medium = base_medium
        sol = model.optimize()
    base_growth = sol.objective_value
    print(base_growth)

    enhancement = {}
    for ex in list(model.exchanges):
        with model:
            medium = model.medium
            base_medium[ex.id] = 10.0
            model.medium = base_medium
            sol = model.optimize()
            if sol.objective_value > base_growth:
                enhancement[ex.id] = sol.objective_value - base_growth
            base_medium.pop(ex.id)

    adds = pd.DataFrame(enhancement.items(), columns=[
                        'exchange', 'diff']).sort_values(by=['diff'], ascending=False)

    return adds



#!/usr/bin/env python
"""Most functions within this module were copied from the MEMOTE GitHub page and modified by Gwendolyn O. Gusak.

This module provides functions to be used to assess the biomass weight as well as normalise it.
"""

import os
import cobra
import logging
from cobra import Reaction
from cobra import Model as cobraModel
from refinegems.io import load_model_libsbml
from six import iteritems
import memote.support.helpers as helpers
from memote.utils import truncate
from typing import Union

__author__ = "MEMOTE and Gwendolyn O. Gusak"


def test_biomass_presence(model: cobraModel) -> Union[list[str], None]:
    """
    Modified from MEMOTE: https://github.com/opencobra/memote/blob/81a55a163262a0e06bfcb036d98e8e551edc3873/src/memote/suite/tests/test_biomass.py#LL42C3-L42C3
    
    Expect the model to contain at least one biomass reaction.

    The biomass composition aka biomass formulation aka biomass reaction
    is a common pseudo-reaction accounting for biomass synthesis in
    constraints-based modelling. It describes the stoichiometry of
    intracellular compounds that are required for cell growth. While this
    reaction may not be relevant to modeling the metabolism of higher
    organisms, it is essential for single-cell modeling.

    Implementation:
    Identifies possible biomass reactions using two principal steps:
    
        1. Return reactions that include the SBO annotation "SBO:0000629" for
        biomass.
        
        2. If no reactions can be identified this way:
        
            1. Look for the ``buzzwords`` "biomass", "growth" and "bof" in reaction IDs.
            2. Look for metabolite IDs or names that contain the ``buzzword`` "biomass" and obtain the set of reactions they are involved in.
            3. Remove boundary reactions from this set.
            4. Return the union of reactions that match the buzzwords and of the reactions that metabolites are involved in that match the buzzword.
        
    This test checks if at least one biomass reaction is present.
    
    If no reaction can be identified return None.

    """
    biomass_rxn = [rxn.id for rxn in helpers.find_biomass_reaction(model)]
    outcome = len(biomass_rxn) > 0
    logging.info(
        """In this model the following {} biomass reaction(s) were
        identified: {}""".format(
            len(biomass_rxn), truncate(biomass_rxn)
        )
    )
    if outcome: return biomass_rxn
    else: return None


def sum_biomass_weight(reaction: Reaction) -> float:
    """
    From MEMOTE: https://github.com/opencobra/memote/blob/81a55a163262a0e06bfcb036d98e8e551edc3873/src/memote/support/biomass.py#L95
    
    Compute the sum of all reaction compounds.

    This function expects all metabolites of the biomass reaction to have
    formula information assigned.

    Parameters
    ----------
    reaction : cobra.core.reaction.Reaction
        The biomass reaction of the model under investigation.

    Returns
    -------
    float
        The molecular weight of the biomass reaction in units of g/mmol.

    """
    return (
        sum(
            -coef * met.formula_weight
            for (met, coef) in iteritems(reaction.metabolites)
        )
        / 1000.0
    )
    
    
def test_biomass_consistency(model: cobraModel, reaction_id: str) -> Union[float, str]:
    """
    Modified from MEMOTE: https://github.com/opencobra/memote/blob/81a55a163262a0e06bfcb036d98e8e551edc3873/src/memote/suite/tests/test_biomass.py#L89
    
    Expect biomass components to sum up to 1 g[CDW].

    This test only yields sensible results if all biomass precursor
    metabolites have chemical formulas assigned to them.
    The molecular weight of the biomass reaction in metabolic models is
    defined to be equal to 1 g/mmol. Conforming to this is essential in order
    to be able to reliably calculate growth yields, to cross-compare models,
    and to obtain valid predictions when simulating microbial consortia. A
    deviation from 1 - 1E-03 to 1 + 1E-06 is accepted.

    Implementation:
    Multiplies the coefficient of each metabolite of the biomass reaction with
    its molecular weight calculated from the formula, then divides the overall
    sum of all the products by 1000.

    """
    reaction = model.reactions.get_by_id(reaction_id)
    try:
        biomass_weight = sum_biomass_weight(reaction)
    except TypeError:
        message = """
        One or more of the biomass components do not have a defined formula or contain unspecified chemical groups.
        The biomass overall weight could thus not be calculated.
        """
        return message
    else:
        if ((1 - 1e-03) < biomass_weight < (1 + 1e-06)):
            logging.info(            
                """The component molar mass of the biomass reaction {} sums up to {}
                which is inside the 1e-03 margin from 1 mmol / g[CDW] / h.
                """.format(
                    reaction_id, biomass_weight
                    )
                )
        else:
            logging.warning(
                """The component molar mass of the biomass reaction {} sums up to {}
                which is outside of the 1e-03 margin from 1 mmol / g[CDW] / h.
                """.format(
                    reaction_id, biomass_weight
                )
            )
    #outcome = (1 - 1e-03) < biomass_weight < (1 + 1e-06) -> Need to implement that for check
    # To account for numerical inaccuracies, a range from 1-1e0-3 to 1+1e-06
    # is implemented in the assertion check
    return biomass_weight


def normalise_biomass(biomass: Reaction, current_sum: float) -> Reaction:
    """Normalises the coefficients according to current biomass weight to one g[CDW]

    Args:
        biomass (Reaction): Biomass function/reaction
        current_sum (float): Biomass weight calculated with sum_biomass_weight in g/mmol

    Returns:
        Reaction: Biomass function/reaction with updated coefficients
    """
    metabs = biomass.metabolites # Get all metabolites
    
    # Normalise & update coefficients
    for (met, coef) in iteritems(metabs):
        metabs[met] = 1/current_sum * coef
    biomass._metabolites = metabs
    
    return biomass

def check_normalise_biomass(model: cobraModel) -> Union[cobraModel, None]:
    """
       1. Checks if at least one biomass reaction is present
       2. For each found biomass reaction checks if it sums up to 1g[CDW]
       3. Normalises the coefficients of each biomass reaction where the sum is not 1g[CDW] until the sum is 1g[CDW]
       4. Returns model with adjusted biomass function(s)
       

    Args:
        model (cobraModel): Model loaded with COBRApy

    Returns:
        cobraModel: COBRApy model with adjusted biomass functions
    """
    
    biomass_rxn = test_biomass_presence(model)
    
    if biomass_rxn:
        for bm_rxn in biomass_rxn:
            bm_weight = test_biomass_consistency(model, bm_rxn)
            if type(bm_weight) == str: logging.error(f'Reaction {bm_rxn}: {bm_weight}')
            else:
                while not ((1 - 1e-03) < bm_weight < (1 + 1e-06)):
                    #normalise_biomass(model.reactions.get_by_id(bm_rxn), bm_weight)
                    model.reactions.get_by_id(bm_rxn).__imul__(1/bm_weight) # -> Maybe add like this?
                    bm_weight = test_biomass_consistency(model, bm_rxn)
                    logging.info(f'For reaction \'{bm_rxn}\' the coefficients changed.')
                
        cobra.io.write_sbml_model(model, f'../{model.id}_tmp.xml')
        model = load_model_libsbml(f'../{model.id}_tmp.xml')
        os.remove(f'../{model.id}_tmp.xml')
            
        return model
    
    else:
        logging.error(f'No biomass objective function was found in the provided model {model.id}.')
        return biomass_rxn



#!/usr/bin/env python
""" Reports mismatches in charges and formulae based on ModelSEED

Extracts ModelSEED data from a given tsv file, extracts all metabolites from a given model. Both lists of metabolites are compared by charge and formula.
"""
import pandas as pd
import re
import numpy as np
from refinegems.io import load_a_table_from_database
from cobra import Model as cobraModel

__author__ = "Famke Baeuerle and Jan-Philipp Leusch"


def get_modelseed_compounds() -> pd.DataFrame:
    """Extracts compounds from ModelSEED which have BiGG Ids

    Returns:
        pd.DataFrame: Table containing ModelSEED data
    """
    # Get only rows where BiGG is contained
    com = load_a_table_from_database("SELECT *, INSTR(aliases, 'BiGG:') bigg FROM modelseed_compounds WHERE bigg > 0")

    def get_bigg_ids(aliases):
        try:
            aliases_list = aliases.split('|')
            bigg = [x[6:] for x in aliases_list if re.search('BiGG: .*', x)]
            return bigg[0]
        except (IndexError, AttributeError):
            return None

    com['BiGG'] = com.apply(lambda row: get_bigg_ids(row['aliases']), axis=1)

    return com.loc[:, ['id', 'name', 'formula', 'mass', 'charge', 'BiGG']]


def get_model_charges(model: cobraModel) -> pd.DataFrame:
    """Extracts all metabolites from model

    Args:
        - model (cobraModel): Model loaded with COBRApy

    Returns:
        pd.DataFrame: Table containing charges and formulae of model metabolites
    """
    charges = {}
    for metab in model.metabolites:
        charges[metab.id[:-2]] = [metab.charge, metab.formula]

    df_charges = pd.DataFrame.from_dict(
        charges,
        orient='index',
        columns=[
            'charge_model',
            'formula_model']).reset_index().rename(
        columns={
            'index': 'BiGG'})

    return df_charges


def get_modelseed_charges(modelseed_compounds: pd.DataFrame) -> pd.DataFrame:
    """Extract table with BiGG, charges and formulae

    Args:
        - modelseed_compounds (pd.DataFrame): ModelSEED data. Output from get_modelseed_compounds.

    Returns:
        pd.DataFrame: Table containing charges and formulae of ModelSEED metabolites
    """
    modelseed_compounds = modelseed_compounds.loc[:, ['charge', 'BiGG', 'formula']].rename(
        columns={'charge': 'charge_modelseed', 'formula': 'formula_modelseed'})
    lst_col = 'BiGG'
    x = modelseed_compounds.assign(
        **{lst_col: modelseed_compounds[lst_col].str.split(';')})
    df_ms = pd.DataFrame({col: np.repeat(x[col].values, x[lst_col].str.len()) for col in x.columns.difference(
        [lst_col])}).assign(**{lst_col: np.concatenate(x[lst_col].values)})[x.columns.tolist()]
    return df_ms


def compare_model_modelseed(model_charges: pd.DataFrame, modelseed_charges: pd.DataFrame) -> pd.DataFrame:
    """Compares tables with charges / formulae from model & modelseed

    Args:
        - model_charges (pd.DataFrame): Charges and formulae of model metabolites. Output of get_model_charges.
        - modelseed_charges (pd.DataFrame): Charges and formulae of ModelSEED metabolites. Output of get_modelseed_charges.

    Returns:
        pd.DataFrame: Table containing info whether charges / formulae match
    """
    df_comp = pd.merge(model_charges, modelseed_charges, on='BiGG', how='left')

    def f(x):
        return True if float(
            x['charge_model']) == x['charge_modelseed'] else False

    def g(x):
        return True if x['formula_model'] == x['formula_modelseed'] else False

    df_comp['charge_match'] = df_comp.apply(f, axis=1)
    df_comp['formula_match'] = df_comp.apply(g, axis=1)

    return df_comp


def get_charge_mismatch(df_comp: pd.DataFrame) -> pd.DataFrame:
    """Extracts metabolites with charge mismatch of model & modelseed

    Args:
        df_comp (pd.DataFrame): Charge and formula mismatches. Output from compare_model_modelseed.

    Returns:
        pd.DataFrame: Table containing metabolites with charge mismatch
    """
    return df_comp.loc[~df_comp['charge_match']].dropna(
        subset=['charge_modelseed'])


def get_formula_mismatch(df_comp: pd.DataFrame) -> pd.DataFrame:
    """Extracts metabolites with formula mismatch of model & modelseed

    Args:
        df_comp (pd.DataFrame): Charge and formula mismatches. Output from compare_model_modelseed.

    Returns:
        pd.DataFrame: Table containing metabolites with formula mismatch
    """
    return df_comp.loc[~df_comp['formula_match']].dropna(
        subset=['formula_modelseed'])


def get_compared_formulae(formula_mismatch: pd.DataFrame) -> pd.DataFrame:
    """Compare formula by atom pattern

    Args:
        formula_mismatch (pd.DataFrame): Table with column containing atom comparison. Output from get_formula_mismatch.

    Returns:
        pd.DataFrame: table containing metabolites with formula mismatch
    """

    def formula_comparison(f1, f2):
        # from Jan Leusch
        formula_pattern = "[A-Z][a-z]?\\d*"
        atom_pattern = '[A-Z][a-z]?'
        atom_number_pattern = '\\d+'
        difference = {}
        f1_dict = {}
        f2_dict = {}
        f1 = re.findall(formula_pattern, f1)
        f2 = re.findall(formula_pattern, f2)
        for p in f1:
            key = re.findall(atom_pattern, p)[0]
            value = re.findall(atom_number_pattern, p)
            if not value:
                value = 1
            else:
                value = (int)(value[0])
            f1_dict[key] = value
        for q in f2:
            key = re.findall(atom_pattern, q)[0]
            value = re.findall(atom_number_pattern, q)
            if not value:
                value = 1
            else:
                value = (int)(value[0])
            f2_dict[key] = value
        difference = f1_dict
        if not f2_dict:
            return difference
        else:
            for q in f2_dict:
                if q in difference:
                    difference[q] -= f2_dict[q]
                else:
                    difference[q] = -f2_dict[q]
        for a in list(difference):
            if difference[a] == 0:
                difference.pop(a)
        return difference

    formula_mismatch['formula_comparison'] = formula_mismatch.apply(
        lambda row: formula_comparison(
            row['formula_model'], row['formula_modelseed']), axis=1)

    return formula_mismatch


def compare_to_modelseed(model: cobraModel) -> tuple[pd.DataFrame, pd.DataFrame]:
    """Executes all steps to compare model metabolites to ModelSEED metabolites

    Args:
        - model (cobraModel): Model loaded with COBRApy

    Returns:
        tuple: Tables with charge (1) & formula (2) mismatches
            (1) pd.DataFrame: Table with charge mismatches 
            (2) pd.DataFrame: Table with formula mismatches
    """
    ms_comp = get_modelseed_compounds()
    model_charges = get_model_charges(model)
    modelseed_charges = get_modelseed_charges(ms_comp)
    df_comp = compare_model_modelseed(model_charges, modelseed_charges)
    charge_mismatch = get_charge_mismatch(df_comp)
    formula_mismatch = get_formula_mismatch(df_comp)
    formula_comp = get_compared_formulae(formula_mismatch)
    return charge_mismatch, formula_comp



#!/usr/bin/env python
""" Provides functions to compare locus tags (genes) found in BioCyc and in the model

Extracts all BioCyc IDs from the annotations and compares them to a list for your organism from BioCyc.
Reactions with BioCyc IDs not found in the model are expanded to a table containing the BioCyc ID,
the locus tag, the protein ID from NCBI, the EC number, the BiGG ID and the KEGG ID. This section needs 
three TXT files from BioCyc as well as the protein FASTA file from the Genbank entry of the organism. 

These three TXT files can be obtained through creating SmartTables in BioCyc and exporting these as 
Spreadsheets with the parameter FrameID (one & two) or Common name (three). 
SmartTable one: Should contain 'Accession-2' and 'Reactions of gene', 
SmartTable two: Should contain 'Reaction', 'Reactants of reaction', 'Products of reaction', 'EC-Number', 'Reaction-Direction' and 'Spontaneous?'
SmartTable three: Should contain 'Compound', 'Chemical Formula' and 'InChI-Key'.
"""
# Get all  possible genes by filtering .gff according to 'bio_type=protein_coding' & 'product=hypothetical protein'
# Compare the list of genes with the ones already in the model & add all missing genes
# Before adding to model check if for all genes that are missing for IMITSC147 identifiers exist
# -> Create tables mapping locus tag to old ID, locus tag to new ID & merge 
# -> Specify user input locus_tag start from NCBI PGAP
from libsbml import Model as libModel
import math
import ast
import numpy as np
import pandas as pd
import libchebipy
import requests
from refinegems.entities import get_model_genes, get_model_reacs_or_metabs, compare_gene_lists
from refinegems.analysis_db import get_bigg2other_db, compare_bigg_model, add_stoichiometric_values_to_reacs, BIGG_METABOLITES_URL
from refinegems.io import parse_fasta_headers

__author__ = "Gwendolyn O. Gusak and Dr. Reihaneh Mostolizadeh"


# Global variable for statistics
statistics_dict = {
   'Missing entity': ['Protein', 'Metabolite', 'Reaction'], 
   'Total': [np.NaN, np.NaN, np.NaN], 
   'Have BiGG ID': [np.NaN, np.NaN, np.NaN], 
   'Can be added': [np.NaN, np.NaN, np.NaN],
   'Notes': [
      'Amount derived from locus tag comparison/Amount remaining for the reactions',
      'Only metabolites that are required for the missing reactions',
      'Only reactions that belong to the missing genes/proteins'
      ]
   }
statistics_df = pd.DataFrame(statistics_dict).set_index('Missing entity')


# Locus tags in GenBank GFF file == BioCyc Accession-2 == Old locus tags in RefSeq GFF file
# Locus tags in RefSeq GFF file == BioCyc Accession-1
# Label in model == Locus tag from GenBank GFF file == BioCyc Accession-2
def get_biocyc_genes2reactions(inpath: str) -> pd.DataFrame:
   """Parses TSV file from BioCyc to retrieve 'Accession-2' & the corresponding 'Reactions of gene'
   
   Args:
      - inpath (str): Path to file from BioCyc containing the 'Accession-2' to 'Reactions of gene' mapping
      
   Returns:
      pd.DataFrame: Table containing only rows where a 'Reaction of gene' exists
   """
   
   biocyc_genes = pd.read_table(inpath, usecols=['Accession-2', 'Reactions of gene'], dtype=str)
   biocyc_genes.rename(columns={'Accession-2': 'locus_tag', 'Reactions of gene': 'Reaction'}, inplace=True)
   biocyc_genes.replace('', np.nan, inplace=True)
   biocyc_genes.dropna(inplace=True)
   return biocyc_genes


def get_missing_genes2reactions(model_libsbml: libModel, inpath: str) -> pd.DataFrame:
   """Retrieves the missing genes and reactions from the BioCyc table according to the 'Accession-2' identifiers

   Args:
      - model_libsbml (libModel):   Model read in with libSBML
      - inpath (str):               Path to file from BioCyc containing the Accession-2 to Reactions of gene mapping
      
   Returns: 
      pd.DataFrame: Table containing only 'Accession-2' & 'Reactions' for the missing genes
   """
   
   gps_in_model = get_model_genes(model_libsbml)
   biocyc_genes = get_biocyc_genes2reactions(inpath)
   missing_biocyc_genes = compare_gene_lists(gps_in_model, biocyc_genes, False)
   
   missing_biocyc_reactions = pd.DataFrame(
      missing_biocyc_genes['Reaction'].str.split('//').tolist(), index=missing_biocyc_genes['locus_tag']
      ).stack()
   missing_biocyc_reactions = missing_biocyc_reactions.reset_index([0, 'locus_tag'])
   missing_biocyc_reactions.columns = ['locus_tag', 'Reaction']
   missing_biocyc_reactions['Reaction'] = missing_biocyc_reactions['Reaction'].str.strip()
   
   # Get amount of missing genes from BioCyc for statistics
   biocyc_genes = missing_biocyc_genes['locus_tag'].unique().tolist()
   statistics_df.loc['Protein', 'Total'] = len(biocyc_genes)
   
   return missing_biocyc_reactions


def get_biocyc_reactions(inpath: str) -> pd.DataFrame:
   """Parses TSV file from BioCyc to retrieve 'Reaction', 'Reactants of reaction', 'Products of reaction', 'EC-Number',
      'Reaction-Direction' & 'Spontaneous?'
   
   Args:
      - inpath (str):   Path to file from BioCyc containing the following columns:
                        'Reaction' 'Reactants of reaction' 'Products of reaction' 'EC-Number' 'Reaction-Direction' 
                        'Spontaneous?'
      
   Returns:
      pd.DataFrame: Table containing all biocyc reactions from provided file
   """
   
   biocyc_reacs = pd.read_table(inpath, usecols=
                                ['Reaction', 'Reactants of reaction', 'Products of reaction', 'EC-Number', 
                                 'Reaction-Direction', 'Spontaneous?'],
                                dtype=str
                                )
   biocyc_reacs.rename(columns=
                       {'Reactants of reaction': 'Reactants', 'Products of reaction': 'Products', 'EC-Number': 'EC'},
                       inplace=True
                       )
   biocyc_reacs.replace('', np.nan, inplace=True)
   biocyc_reacs['Spontaneous?'] = biocyc_reacs['Spontaneous?'].fillna('F')
   biocyc_reacs.dropna(subset=['Reaction', 'Reactants', 'Products', 'EC'], inplace=True)
   return biocyc_reacs


def extract_metabolites_from_reactions(missing_reactions: pd.DataFrame) -> tuple[pd.DataFrame, pd.DataFrame]:
   """Extracts a set of all reactants & products from the missing reactions
   
   Args:
      - missing_reactions (pd.DataFrame): Table containing all missing reactions found through the 
                                       missing genes
                                          
   Returns:
      tuple: Two tables (1) & (2) 
         (1) pd.DataFrame: Table with the column Compound containing all compounds required for the missing BioCyc reactions
         (2) pd.DataFrame: Table with the column bigg_id containing all compounds required for the missing BiGG reactions
   """
   # Get all BioCyc metabolites necessary for the BioCyc reactions
   biocyc_reactants = [r for row in missing_reactions['Reactants'] for r in row]
   biocyc_products = [p for row in missing_reactions['Products'] for p in row]
   biocyc_metabolites = list(set([*biocyc_reactants, *biocyc_products]))
   
   # Get all BiGG metabolites necessary for the BiGG reactions
   bigg_reactions = [ast.literal_eval(str(reac_dict)) for reac_dict in missing_reactions['bigg_reaction']]
   bigg_reactants = [r for reac_dict in bigg_reactions for r in reac_dict.get('reactants')]
   bigg_products = [p for reac_dict in bigg_reactions for p in reac_dict.get('products')]
   bigg_metabolites = list(set([*bigg_reactants, *bigg_products]))
   
   return (pd.DataFrame(biocyc_metabolites, columns=['Compound']), pd.DataFrame(bigg_metabolites, columns=['bigg_id']))


def get_missing_reactions(
   model_libsbml: libModel, genes2reaction: pd.DataFrame, inpath: str
   ) -> tuple[tuple[pd.DataFrame, pd.DataFrame], pd.DataFrame]:
   """Subsets the BioCyc table with the following columns: 
   'Reaction' 'Reactants of reaction' 'Products of reaction' 'EC-Number' 'Reaction-Direction' 'Spontaneous?'
      to obtain the missing reactions with all the corresponding data 
      & Adds the according BiGG Reaction identifiers

   Args:
      - model_libsbml (libModel):      Model read in with libSBML
      - genes2reaction (pd.DataFrame): Table containing only 'Accession-2' & 'Reactions' for the 
                                       missing genes
      - inpath (str):                  Path to file from BioCyc containing the following columns:
                                       'Reaction' 'Reactants of reaction' 'Products of reaction' 'EC-Number' 
                                       'Reaction-Direction' 'Spontaneous?'
      
   Returns:
      tuple: Tuple (1) & table (2)
         (1) tuple: Two tables (1) & (2)
            (1) pd.DataFrame: Table containing only the metabolites corresponding to the missing BioCyc reactions
            (2) pd.DataFrame: Table containing only the metabolites corresponding to the missing BiGG reactions
         (2) pd.DataFrame: Table containing the missing reactions with the corresponding data
   """
   model_reacs = get_model_reacs_or_metabs(model_libsbml)
   biocyc_reacs = get_biocyc_reactions(inpath)
   
   # Get missing reactions from missing genes
   missing_reactions = genes2reaction.merge(biocyc_reacs, on='Reaction')
   
   # Turn entries with '//' into lists
   missing_reactions['Reactants'] = missing_reactions['Reactants'].str.split('\s*//\s*')
   missing_reactions['Products'] = missing_reactions['Products'].str.split('\s*//\s*')
   missing_reactions['EC'] = missing_reactions['EC'].str.split('\s*//\s*')
   
   # Turn locus_tag column into lists of locus tags per reaction
   locus_tags_as_list = missing_reactions.groupby('Reaction')['locus_tag'].apply(list).reset_index(name='locus_tag')
   missing_reactions.drop('locus_tag', axis=1, inplace=True)
   missing_reactions = locus_tags_as_list.merge(missing_reactions, on='Reaction')
   statistics_df.loc['Reaction', 'Total'] = len(missing_reactions['Reaction'].unique().tolist())
   
   # Get BiGG BioCyc
   bigg2biocyc_reacs = get_bigg2other_db('BioCyc')
   
   # Subset missing_reactions with BiGG BioCyc
   missing_reactions.rename(columns={'Reaction': 'BioCyc'}, inplace=True)
   missing_reactions = bigg2biocyc_reacs.merge(missing_reactions, on='BioCyc')
   
   # Get amount of missing reactions that have a BiGG ID
   statistics_df.loc['Reaction', 'Have BiGG ID'] = len(missing_reactions['BioCyc'].unique().tolist())
   
   # Subset missing_reactions with model_reacs
   missing_reactions = compare_bigg_model(missing_reactions, model_reacs)
   
   # Get amount of missing reactions that are not in the model
   statistics_df.loc['Reaction', 'Can be added'] = len(missing_reactions['bigg_id'].unique().tolist())
   
   # Add reactants & products dictionary with stoichiometric values to the reactions table
   missing_reactions = add_stoichiometric_values_to_reacs(missing_reactions)
   
   # Get all metabolites for the missing reactions
   biocyc_metabs_from_reacs, bigg_metabs_from_reacs = extract_metabolites_from_reactions(missing_reactions)
   return (biocyc_metabs_from_reacs, bigg_metabs_from_reacs), missing_reactions 


def get_biocyc_metabolites(inpath: str) -> pd.DataFrame:
   """Parses TSV file from BioCyc to retrieve 'Compound (Object ID)' 'Chemical Formula' 'InChI-Key' 'ChEBI'
   
   Args:
      - inpath (str):   Path to file from BioCyc containing the following columns: 
                        'Compound' 'Object ID' 'Chemical Formula' 'InChI-Key' 'ChEBI'
      
   Returns:
      pd.DataFrame: Table containing all biocyc metabolites from provided file
   """
   
   biocyc_metabs = pd.read_table(inpath, usecols=['Object ID', 'Chemical Formula', 'InChI-Key', 'ChEBI'], dtype=str)
   biocyc_metabs.rename(columns={'Object ID': 'Compound'}, inplace=True)
   biocyc_metabs.replace('', np.nan, inplace=True)
   biocyc_metabs.dropna(inplace=True)
   return biocyc_metabs

   
def get_missing_metabolites(
   model_libsbml: libModel, metabs_from_reacs: tuple[pd.DataFrame, pd.DataFrame], inpath: str
   ) -> pd.DataFrame: 
   """Subsets the BioCyc table with the following columns: 'Compound' 'Chemical Formula' 'InChI-Key' 'ChEBI'
      to obtain the missing metabolites with all the corresponding data
      & Adds the according BiGG Compound identifiers
      
   Args:
      - model_libsml (libModel):          Model read in with libSBML 
      - metabs_from_reacs (tuple): Two tables containing only the metabolites corresponding to the 
                                          missing reactions for either BioCyc (1) or BiGG (2)
      - inpath (str):                     Path to file from BioCyc containing the following columns:
                                          'Compound' 'Chemical Formula' 'InChI-Key'
      
   Returns:
      tuple: Two tables (1) & (2)
         (1): Table containing the metabolites corresponding to the missing reactions without BiGG IDs
         (2): Table containing the metabolites corresponding to the missing reactions with BiGG IDs
   """
   model_metabs = get_model_reacs_or_metabs(model_libsbml, True)
   biocyc_metabs = get_biocyc_metabolites(inpath)
   biocyc_metabs.rename(columns={'Compound': 'BioCyc'}, inplace=True)
   biocyc_metabs_from_reacs, bigg_metabs_from_reacs = metabs_from_reacs
   
   # Get amount of missing BioCyc metabolites for all missing reactions
   statistics_df.loc['Metabolite', 'Total'] = len(biocyc_metabs_from_reacs['Compound'].unique().tolist())
   
   # Get BiGG BioCyc
   bigg2biocyc_metabs = get_bigg2other_db('BioCyc', True)
   
   # Subset biocyc_metabs with BiGG BioCyc -> To get only metabolites with BiGG IDs
   missing_metabolites = bigg2biocyc_metabs.merge(biocyc_metabs, on='BioCyc') # missing_metabolites
   
   # Filter for all required BiGG metabolites for the missing BiGG reactions
   missing_metabolites = bigg_metabs_from_reacs.merge(missing_metabolites, how='left', on='bigg_id')
   
   # Get amount of missing BioCyc metabolites that have a BiGG ID
   missing_biocyc_metabs = missing_metabolites.dropna(subset=['BioCyc'])
   statistics_df.loc['Metabolite', 'Have BiGG ID'] = len(missing_biocyc_metabs['BioCyc'].unique().tolist())
   
   # Subset missing_metabolites with model_metabs
   missing_metabolites = compare_bigg_model(missing_metabolites, model_metabs, True)
   
   # Get amount of missing metabolites that can & should be added to the model
   statistics_df.loc['Metabolite', 'Can be added'] = len(missing_metabolites['bigg_id'].unique().tolist())
   return missing_metabolites


def get_missing_genes(missing_reactions: pd.DataFrame, fasta: str) -> tuple[pd.DataFrame, pd.DataFrame]:
   """Retrieves all missing genes that belong to the obtained missing reactions
   
   Args: 
      - missing_reactions (pd.DataFrame): Table containing all obtained missing reactions
      - fasta (str):                   Path to a FASTA file where the headers contain the information protein_id and locus_tag
         
   Returns:
      tuple: Two tables (1) & (2)
         (1) pd.DataFrame: Table with the columns locus_tag, Protein_id & Model_id 
                           (The model_id is similar to how CarveMe generates the GeneProduct ID.)
         (2) pd.DataFrame: The input pandas dataframe for the reactions where column 'locus_tag' is exchanged by 'gene_product'
   """
   # Get locus tags from the missing reactions
   locus_tags = list(set([lt for row in missing_reactions['locus_tag'] for lt in row]))
   locus_tags_df = pd.DataFrame(pd.Series(locus_tags), columns=['locus_tag'])
   
   # Get protein and GeneProduct ID for the model from FASTA file
   ids_df = parse_fasta_headers(fasta, id_for_model=True)
   
   # Get the complete dataframe with the protein & model id
   missing_genes = locus_tags_df.merge(ids_df, on='locus_tag')
   statistics_df.loc['Protein', 'Can be added'] = len(missing_genes['locus_tag'].unique().tolist())
   
   # Replace the locus tags in the reaction dataframe with the gene model ID
   def transform_lt_into_gp_model_id(locus_tag_list: list[str]) -> list[str]:
      return [missing_genes.loc[lt, 'model_id'] for lt in locus_tag_list]
   
   missing_genes.set_index('locus_tag', inplace=True)
   missing_reactions['gene_product'] = missing_reactions['locus_tag'].map(transform_lt_into_gp_model_id)
   missing_genes.reset_index(inplace=True)
   missing_reactions.drop('locus_tag', axis=1, inplace=True)
   
   return missing_genes, missing_reactions


def add_charges_chemical_formulae_to_metabs(missing_metabs: pd.DataFrame) -> pd.DataFrame:
   """Adds charges & chemical formulae from CHEBI/BiGG to the provided dataframe

   Args:
      - missing_metabs (pd.DataFrame): Table containing metabolites & the respective CHEBI & BiGG IDs
         
   Returns:
      pd.DataFrame: Input table extended with the charges & chemical formulas obtained from CHEBI/BiGG
   """
   
   # Finds the charges through the ChEBI/BiGG API, defaults to: 0
   def find_charge(row: pd.Series) -> int:
      chebi_id = str(int(row.get('ChEBI'))) if not math.isnan(float(row.get('ChEBI'))) else None
      bigg_id = str(row.get('bigg_id'))
      charge = None
      if chebi_id:  # Get charge from ChEBI (Returns always a charge)
         chebi_entity = libchebipy.ChebiEntity('CHEBI:' + chebi_id)
         return chebi_entity.get_charge()
      elif bigg_id != 'nan':  # Get charge from BiGG if no ChEBI ID available
         try:
            charge = requests.get(BIGG_METABOLITES_URL + bigg_id[:-2]).json()['charges'][0]  # Take first charge
         except ValueError:
            pass   
         # If no charge was found, charge=0
         return charge if charge else 0
   
   # Finds the chemical formula through the ChEBI/BiGG API, defaults to: 'No formula'
   def find_formula(row: pd.Series) -> str:
      chebi_id = str(int(row.get('ChEBI'))) if not math.isnan(float(row.get('ChEBI'))) else None
      bigg_id, chem_form = str(row.get('bigg_id')), str(row.get('Chemical Formula'))
      chem_formula = None
      if chebi_id: # Get formula from ChEBI
         chebi_entity = libchebipy.ChebiEntity('CHEBI:' + chebi_id)
         chem_formula = chebi_entity.get_formula()
      if not chem_formula:  # If no formula was found with ChEBI/No ChEBI ID available
         if bigg_id != 'nan': # Get formula from BiGG
            try:
               chem_formula = requests.get(BIGG_METABOLITES_URL + bigg_id[:-2]).json()['formulae'][0]  # Take first formula
            except ValueError:
               pass
         if not chem_formula: # If no formula was found with BiGG ID
            # Get formula already existing in dataframe or set to 'No formula'
            chem_formula = chem_form if chem_form != 'nan' else 'No formula'
      return chem_formula
   
   missing_metabs['charge'] = missing_metabs.apply(find_charge, axis=1)
   missing_metabs['New Chemical Formula'] = missing_metabs.apply(find_formula, axis=1)
   missing_metabs['Chemical Formula'] = missing_metabs['New Chemical Formula']
   missing_metabs.drop('New Chemical Formula', axis=1, inplace=True)
   
   return missing_metabs

# Inspired by Dr. Reihaneh Mostolizadeh's function to add BioCyc reactions to a model
def replace_reaction_direction_with_fluxes(missing_reacs: pd.DataFrame) -> pd.DataFrame:
   """Extracts the flux lower & upper bounds for each reaction through the entries in column 'Reaction-Direction'
   
   Args:
      - missing_reacs (pd.DataFrame): Table containing reactions & the respective Reaction-Directions
         
   Returns:
      pd.DataFrame: Input table extended with the fluxes lower & upper bounds obtained from 
                     the Reaction-Directions
   """
    
   def get_fluxes(row: pd.Series) -> dict[str: str]:
      direction = row['Reaction-Direction']
      fluxes = {}
      
      if type(direction) == float:
         # Use default bounds as described in readthedocs from COBRApy
         fluxes['lower_bound'] = 'cobra_0_bound'
         fluxes['upper_bound'] = 'cobra_default_ub'
      elif 'RIGHT-TO-LEFT' in direction:
         fluxes['lower_bound'] = 'cobra_default_lb'
         fluxes['upper_bound'] = 'cobra_0_bound'
      elif 'LEFT-TO-RIGHT' in direction:
         fluxes['lower_bound'] = 'cobra_0_bound'
         fluxes['upper_bound'] = 'cobra_default_ub'
      elif 'REVERSIBLE' in direction:
         fluxes['lower_bound'] = 'cobra_default_lb'
         fluxes['upper_bound'] = 'cobra_default_ub'
      
      return str(fluxes)
   
   missing_reacs['fluxes'] = missing_reacs.apply(get_fluxes, axis=1)
   missing_reacs.drop('Reaction-Direction', axis=1, inplace=True)
   
   return missing_reacs


def biocyc_gene_comp(
   model_libsbml: libModel, biocyc_file_paths: list[str]
   ) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
   """Main function to retrieve the tables for missing genes, metabolites and reactions from BioCyc
   
   Args:
      - model_libsbml (libModel):   libSBML Model object
      - biocyc_file_paths (list):   List of the files required for the BioCyc analysis
         
   Returns: 
      tuple: Five tables (1) - (5)
         (1) pd.DataFrame: Table containing the statistics of the BioCyc gapfill analysis
         (2) pd.DataFrame: Table containing the missing genes that belong to the missing reactions
         (3) pd.DataFrame: Table containing the missing metabolites with BiGG IDs belonging to the missing reactions
         (4) pd.DataFrame: Table containing the missing metabolites without BiGG IDs belonging to the missing reactions
         (5) pd.DataFrame: Table containing the missing reactions
   """
   # Extract missing reactions from all missing genes
   genes2reactions = get_missing_genes2reactions(model_libsbml, biocyc_file_paths[0])
   metabs_from_reacs, missing_reactions_df = get_missing_reactions(model_libsbml, genes2reactions, biocyc_file_paths[1])
   missing_reactions_df = replace_reaction_direction_with_fluxes(missing_reactions_df)
   
   # Extract missing metabolites that belong to the missing reactions
   missing_metabolites_df= get_missing_metabolites(model_libsbml, metabs_from_reacs, biocyc_file_paths[2]) 
   missing_metabolites_df = add_charges_chemical_formulae_to_metabs(missing_metabolites_df)
   
   # Extract missing genes that belong to the missing reactions
   missing_genes_df, missing_reactions_df = get_missing_genes(missing_reactions_df, biocyc_file_paths[3])
   
   # Remove index from statistics_df
   statistics_df.reset_index(inplace=True)
   
   return (statistics_df, missing_genes_df, missing_metabolites_df, missing_reactions_df) 
   


#!/usr/bin/env python
""" Provides functions to compare and visualize multiple models

Can mainly be used to compare growth behaviour of multiple models. All other stats are shown in the memote report.
"""

import pandas as pd
import matplotlib
import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
from tqdm import tqdm
from venn import venn
from libsbml import Model as libModel
from cobra import Model as cobraModel
from refinegems.io import load_medium_from_db, search_sbo_label
from refinegems.growth import growth_one_medium_from_default, growth_one_medium_from_minimal
from refinegems.investigate import initial_analysis, get_reactions_per_sbo

__author__ = "Famke Baeuerle"

def plot_initial_analysis(models: list[libModel]):
    """Creates bar plot of number of entities per Model

    Args:
        - models (list[libModel]): Models loaded with libSBML

    Returns:
        plot: Pandas Barchart
    """
    numbers = pd.DataFrame([initial_analysis(model) for model in models], columns=['model', 'metabolites', 'reactions', 'genes'])
    ax = numbers.set_index('model').plot.bar(y=['metabolites', 'reactions', 'genes'], figsize=(8, 5), cmap='Paired', rot=0)
    # commented is possibility to integrate memote scores
    #numbers.set_index('model').plot(y='Memote score', ax=ax, use_index=False, linestyle=':', secondary_y='Memote score', color='k', marker='D', legend=True)
    #ax.right_ax.set_ylabel('Memote score [%]')
    #ax.right_ax.legend(loc='upper right', bbox_to_anchor=[0.98, 0.9])
    #ax.right_ax.set_ylim([75, 95])
    ax.legend(title=False, loc='upper left', ncol=3, frameon=False)
    ylim = numbers.drop('model', axis=1).max().max() + 200
    ax.set_ylim([0,ylim])
    ax.set_xlabel('')
    ax.tick_params(axis='x',which='both', bottom=False,top=False)
    return ax

def get_sbo_mapping_multiple(models: list[libModel]) -> pd.DataFrame:
    """Determines number of reactions per SBO Term and adds label of SBO Terms

    Args:
        - models (list[libModel]): Models loaded with libSBML

    Returns:
        pd.DataFrame: SBO Terms, number of reactions per Model and SBO Label
    """
    mappings = {}
    for model in models:
        mappings[model.id] = get_reactions_per_sbo(model)
    df = pd.DataFrame.from_dict(mappings)
    df = df.reset_index().rename({'index': 'SBO-Term'}, axis=1)
    df['SBO-Name'] = df['SBO-Term'].apply(search_sbo_label)
    return df

def plot_rea_sbo_multiple(models: list[libModel], rename=None):
    """Plots reactions per SBO Term in horizontal bar chart with stacked bars for the models

    Args:
        - models (list[libModel]): Models loaded with libSBML
        - rename (dict, optional): Rename model ids to custom names. Defaults to None.

    Returns:
        plot: Pandas stacked barchart
    """
    map = get_sbo_mapping_multiple(models)
    id_list = [mod.id for mod in models]
    map = map[(map[id_list]>3).all(axis=1)]
    map = map.drop('SBO-Term', axis=1).sort_values(id_list[0]).set_index('SBO-Name')
    if rename is not None:
        map = map.rename(rename, axis=1)
    fig = map.plot.barh(stacked=True, width=.8, figsize=(8,10))
    fig.set_ylabel('')
    fig.set_xlabel('number of reactions', fontsize=16)
    fig.legend(loc='lower right')
    return fig

def plot_venn(models: list[cobraModel], entity: str, perc: bool=False, rename=None):
    """Creates Venn diagram to show the overlap of model entities

    Args:
        - models (list[cobraModel]): Models loaded with cobrapy
        - entity (str): Compare on metabolite|reaction
        - perc (bool, optional): True if percentages should be used. Defaults to False.
        - rename (dict, optional): Rename model ids to custom names. Defaults to None.

    Returns:
        plot: Venn diagram 
    """
    intersec = {}
    for model in models:
        reas = []
        if entity == 'metabolite':
            for rea in model.metabolites:
                reas.append(rea.id)
        if entity == 'reaction':
            for rea in model.reactions:
                reas.append(rea.id)
        if rename is not None:
            intersec[rename[model.id]] = set(reas)
        else:
            intersec[model.id] = set(reas)
    if perc:
        fig = venn(intersec, fmt="{percentage:.1f}%")
    else:
        fig = venn(intersec)
    return fig

def plot_heatmap_dt(growth: pd.DataFrame):
    """Creates heatmap of simulated doubling times with additives
    
    Args:
        - growth (pd.DataFrame): Containing growth data from simulate_all
        
    Returns:
        plot: Seaborn Heatmap
    """
    growth=growth.set_index(['medium', 'model']).sort_index().T.stack()
    growth.columns.name=None
    growth.index.names = (None,None)
    growth.index.name=None
    growth.index = growth.index.get_level_values(1)
    growth[growth > 500] = 0
    growth[growth < 0] = 0
    growth.replace([np.inf, -np.inf], 0, inplace=True)
    over_growth = growth.max().max() + 6
    growth.replace(np.nan, over_growth, inplace=True)
    under_growth = growth.min().min() - 5
    vmin= under_growth if under_growth > 1e-5 else 1e-5 #Use same threshhold as in find_missing_essential in growth
    vmax=over_growth - 1
    annot = growth.copy()
    annot = annot.round().astype(int)
    annot[annot < 1e-5] = ''
    annot.replace(over_growth.round().astype(int), 'No data', inplace=True)
    cmap=matplotlib.cm.get_cmap('YlGn').copy()
    cmap.set_under('black')
    cmap.set_over('white')
    fig, ax = plt.subplots(figsize=(10,8))
    sns.heatmap(growth.T, 
                annot=annot.T, 
                annot_kws={"fontsize":15},
                vmin=vmin, 
                vmax=vmax,
                cmap=cmap, 
                linewidth=.5, 
                cbar_kws = {'orientation':'vertical', 'label':'doubling time [min]', 'extend': 'min', 'extendrect':True},
                ax=ax,
                fmt=''
                )
    rotation = 40 if len(growth.index) > 3 else 0
    plt.tick_params(rotation=0, bottom=False, top=False, left=False, right=False)
    ax.set_xticklabels(ax.get_xticklabels(), rotation=rotation, ha="right")
    return fig

def plot_heatmap_native(growth: pd.DataFrame):
    """Creates a plot were if growth without additives is possible is marked from yellow to green otherwise black

    Args:
        - growth (pd.DataFrame): Containing growth data from simulate_all
        
    Returns:
        plot: Seaborn Heatmap
    """
    def get_native_growth(row):
        if row['complete'] == True:
            return row['doubling_time [min]']
        else:
            return 0
    
    growth['native_growth'] = growth.apply(get_native_growth, axis=1)
    growth = growth[['medium', 'model', 'native_growth']]
    growth=growth.set_index(['medium', 'model']).sort_index().T.stack()
    growth.columns.name=None
    growth.index.names = (None,None)
    growth.index.name=None
    growth.index = growth.index.get_level_values(1)
    growth[growth > 500] = 0
    growth[growth < 0] = 0
    growth.replace([np.inf, -np.inf], 0, inplace=True)
    over_growth = growth.max().max() + 6
    growth.replace(np.nan, over_growth, inplace=True)
    annot = growth.copy()
    annot = annot.round().astype(int)
    annot[annot == np.nan] = 'No data'
    annot[annot < 1e-5] = ''
    annot.replace(over_growth.round().astype(int), 'No data', inplace=True)
    under_growth = growth.min().min() - 5
    vmin= under_growth if under_growth > 1e-5 else 1e-5 #Use same threshhold as in find_missing_essential in growth
    vmax= over_growth - 1
    cmap=matplotlib.cm.get_cmap('YlGn').copy()
    cmap.set_under('black')
    cmap.set_over('white')
    fig, ax = plt.subplots(figsize=(10,8))
    sns.heatmap(growth.T,
                annot=annot.T, 
                annot_kws={"fontsize":15},
                vmin=vmin,
                vmax=vmax,
                cmap=cmap,
                linewidth=.5, 
                ax=ax,
                cbar_kws={'orientation':'vertical', 'label':'doubling time [min]', 'extend': 'min', 'extendrect':True},
                fmt=''
                )
    plt.xticks(rotation=0)
    plt.yticks(rotation=0)
    rotation = 40 if len(growth.index) > 3 else 0
    plt.tick_params(rotation=0, bottom=False, top=False, left=False, right=False)
    ax.set_xticklabels(ax.get_xticklabels(), rotation=rotation, ha="right")
    return fig

def simulate_all(models: list[cobraModel], media: list[str], basis: str, anaerobic: bool) -> pd.DataFrame:
    """Does a run of growth simulation for multiple models on different media

    Args:
        - models (list[cobraModel]): Models loaded with cobrapy
        - media (list[str]): Media of interest (f.ex. LB, M9, ...)
        - basis (str): Either default_uptake (adding metabs from default) or minimal_uptake (adding metabs from minimal medium)
        - anaerobic (bool): If True 'EX_o2_e' is set to 0.0 to simulate anaerobic conditions

    Returns:
        pd.DataFrame: table containing the results of the growth simulation
    """
    growth = pd.DataFrame()
    for medium_id in tqdm(media):
        medium = load_medium_from_db(medium_id)
        for model in models:
            essentials_given = False
            if (basis=='default_uptake'):
                growth_one = growth_one_medium_from_default(model, medium, anaerobic).drop('missing exchanges', axis=1)
            elif (basis == 'minimal_uptake'):
                growth_one = growth_one_medium_from_minimal(model, medium, anaerobic).drop('missing exchanges', axis=1)
            if growth_one['essential'].dropna().size == 0:
                essentials_given = True
            else:
                growth_list = growth_one['essential'].dropna().to_list()
                growth_string = ', '.join(growth_list)
                essentials_given = growth_string
            growth_one = growth_one.drop('essential', axis=1)
            growth_one['complete'] = essentials_given
            growth_one = growth_one.dropna()
            growth_one['model'] = model.id
            growth_one = growth_one[['model', 'medium', 'doubling_time [min]', 'growth_value', 'complete']]
            growth_one['doubling_time [min]'].astype(float).round(2)
            growth_one['growth_value'].astype(float).round(2)
            growth = growth.append(
                growth_one, 
                ignore_index=True)

    return growth




#!/usr/bin/env python
"""The `gapfill` module can be used either with KEGG were you only need the KEGG organism ID or with BioCyc or with both (Options: 'KEGG', 'BioCyc', 'KEGG+BioCyc').
    For how to obtain the BioCyc tables look into the documentation under 'Filling gaps with refineGEMs' > 'Automated gap filling'.
    
    Run times: 
    
        * 'KEGG': ~ 2h
        * 'BioCyc': ~ 45mins - 1h
        * 'KEGG+BioCyc': ~ 3 - 4h  
"""
import ast
import math
from libsbml import Model as libModel
import refinegems.analysis_kegg as rga_kegg
import refinegems.analysis_biocyc as rga_biocyc
from refinegems.curate import update_annotations_from_others
from refinegems.cvterms import add_cv_term_metabolites, add_cv_term_reactions 
from refinegems.entities import create_gp, create_species, create_reaction
import pandas as pd
import numpy as np
from typing import Union
from colorama import init as colorama_init
from colorama import Fore

__author__ = "Famke Baeuerle and Gwendolyn O. Gusak"

'''Skeleton for functions that could be used for a lab strain/organism which is in no database contained
def get_genes_from_gff():
    pass


def get_related_metabs_reactions_blast():
    pass


def gff_gene_comp():
    pass
'''


def gap_analysis(model_libsbml: libModel, gapfill_params: dict[str: str], filename: str) -> Union[pd.DataFrame, tuple]:  # (Genbank) GFF file
    """| Main function to infer gaps in a model by comparing the locus tags of the GeneProducts 
       | to KEGG/BioCyc/both

    Args:
        - model_libsbml (libModel): Model loaded with libSBML
        - gapfill_params (dict): Dictionary obtained from YAML file containing the parameter mappings 
        - filename (str): Path to output file for gapfill analysis result
        
    Returns:
        - Case 'KEGG'
            pd.DataFrame: Table containing the columns 'bigg_id' 'locus_tag' 'EC' 'KEGG' 'name' 'GPR'
        - Case 'BioCyc'
            tuple: Five tables (1) - (4)
                (1) pd.DataFrame: Gap fill statistics with the columns 
                                    'Missing entity' 'Total' 'Have BiGG ID' 'Can be added' 'Notes'
                (2) pd.DataFrame: Genes with the columns 
                                    'locus_tag' 'protein_id' 'model_id' 'name'
                (3) pd.DataFrame: Metabolites with the columns 
                                    'bigg_id' 'name' 'BioCyc' 'compartment' 'Chemical Formula' 'InChI-Key' 'ChEBI' 'charge'  
                (4) pd.DataFrame: Reactions with the columns 
                                    'bigg_id' 'name' 'BioCyc' 'locus_tag' 'Reactants' 'Products' 'EC' 'Fluxes' 'Spontaneous?' 
                                    'bigg_reaction'
                
        - Case 'KEGG+BioCyc': 
            tuple: Five tables (1)-(4) from output of 'BioCyc' & (5) from output of 'KEGG'
                    -> Table reactions contains additionally column 'KEGG'
    """
    colorama_init(autoreset=True)
    db_to_compare = gapfill_params['db_to_compare']
    result = None
    
    if db_to_compare not in ['KEGG', 'BioCyc', 'KEGG+BioCyc']:  # 'GFF', 
        print(f'{Fore.RED}To use the module gapfill the parameter of db_to_compare has to be set to one of the following' 
              + ' options:\n- \'KEGG\'\n- \'BioCyc\'\n- \'KEGG+BioCyc\'\nAdditionally, the required parameters' 
              + ' for each option need to be specified.\n- \'biggreactions\' and \'gapfill\' are required for all options.'
              + '\n- \'organismid\' is required only for the options \'KEGG\' and \'KEGG+BioCyc\'.\n- \'biocyc_tables\'' 
              + ' is only required for the options \'BioCyc\' and \'KEGG+BioCyc\'.')  # \n- \'GFF\'
        return
   
    if db_to_compare == 'KEGG':
        if gapfill_params['organismid']:
            missing_kegg = rga_kegg.kegg_gene_comp(model_libsbml, 
                                                   gapfill_params['organismid'], 
                                                   gapfill_params['gff_file']
                                                   )
            result = missing_kegg
        else:
            print(f'{Fore.RED}To use the KEGG comparison the specification of the organismid is obligatory.\n' +
                  'If there is no organismid available for your organism in KEGG but an entry for your organism exists in BioCyc, use the option \'BioCyc\'.\n' +
                  'If no entry for your organism exists in KEGG and/or BioCyc, the gap analysis cannot be done.')
            # use one of the options \'BioCyc\' or \'GFF\'
        
    elif db_to_compare == 'BioCyc':
        missing_biocyc = rga_biocyc.biocyc_gene_comp(model_libsbml, 
                                                     gapfill_params['biocyc_files']
                                                     )
        result = missing_biocyc
        
        ''' Implement here call of function that can be used with lab strain/organism which is in no database contained
        elif db_to_compare == 'GFF':
            gff_genes = gff_gene_comp(model_libsbml, 
                                    gapfill_params['gff_file']
                                    )
            result = gff_genes
        '''
        
    elif db_to_compare == 'KEGG+BioCyc':
        missing_kegg_reacs = rga_kegg.kegg_gene_comp(model_libsbml, 
                                               gapfill_params['organismid'], 
                                               gapfill_params['gff_file']
                                               )
        missing_kegg_reacs.drop(['name', 'locus_tag', 'EC'], axis=1, inplace=True)
        missing_biocyc = rga_biocyc.biocyc_gene_comp(model_libsbml, 
                                                     gapfill_params['biocyc_files']
                                                     )
        stats, missing_biocyc_genes, missing_biocyc_metabs, missing_biocyc_reacs = missing_biocyc 
        missing_combined_reacs = missing_biocyc_reacs.merge(missing_kegg_reacs[['bigg_id', 'KEGG']], how='left', on='bigg_id')
        result = (stats, missing_biocyc_genes, missing_biocyc_metabs, missing_combined_reacs, missing_kegg_reacs)
        
        
    if type(result) == tuple:
        with pd.ExcelWriter(f'{filename}.xlsx') as writer:
            result[0].to_excel(writer, sheet_name='gap fill statistics', index=False)
            result[1].to_excel(writer, sheet_name='genes', index=False)
            result[2].to_excel(writer, sheet_name='metabolites', index=False)
            result[3].to_excel(writer, sheet_name='reactions', index=False)
            if len(result) == 5:
                result[4].to_excel(writer, sheet_name='KEGG reactions', index=False)
    else:
        with pd.ExcelWriter(f'{filename}.xlsx') as writer:
            result.to_excel(writer, sheet_name='KEGG reactions', index=False)
        
    return result
    
    
def gapfill_model(model_libsbml: libModel, gap_analysis_result: Union[str, tuple]) -> libModel:
    """Main function to fill gaps in a model from a table

    Args:
        - model_libsbml (libModel): Model loaded with libSBML
        - gap_analysis_result (str|tuple): Path to Excel file from gap_analysis|Tuple of pd.DataFrames obtained from gap_analysis
        
    Returns:
        libModel: Gap filled model
    """
    model = model_libsbml
    missing_genes_df, missing_metabs_df, missing_reacs_df = None, None, None
    if type(gap_analysis_result) == tuple:  # Tuple of pandas dataframes from gap_analysis
        missing_genes_df = gap_analysis_result[1]
        missing_metabs_df = gap_analysis_result[2]
        missing_reacs_df = gap_analysis_result[3]
    else:  # Excel file from user-input
        with pd.ExcelFile(gap_analysis_result) as reader:
            gp_analysis_res = pd.read_excel(reader, sheet_name=['genes', 'metabolites', 'reactions'])
        missing_genes_df = gp_analysis_res.get('genes')
        missing_metabs_df = gp_analysis_res.get('metabolites')
        missing_reacs_df = gp_analysis_res.get('reactions')
        
    missing_genes_df = missing_genes_df.replace(np.nan, None)
    missing_metabs_df = missing_metabs_df.replace(np.nan, None)
    missing_reacs_df = missing_reacs_df.replace(np.nan, None)
    
    # (1) Add all missing genes needed for the missing reactions
    for _, row in missing_genes_df.iterrows():
        gp, model = create_gp(model_libsbml, row['model_id'], row['name'], row['locus_tag'], row['protein_id'])
    
    # (2) Add all missing metabolites needed for the missing reactions
    for _, row in missing_metabs_df.iterrows():
        sp, model = create_species(model_libsbml, row['bigg_id'], row['name'], row['compartment'], row['charge'], row['Chemical Formula'])
        if 'BioCyc' in missing_metabs_df.columns:
            biocyc_row = ast.literal_eval(str(row['BioCyc']).replace('nan', 'None'))
            if biocyc_row:
                for biocyc_id in biocyc_row:
                    if biocyc_id:
                        add_cv_term_metabolites(biocyc_id, 'BioCyc', sp)
                        add_cv_term_metabolites(biocyc_id, 'METACYC', sp)
                
        if 'InChI-Key' in missing_metabs_df.columns:
            inchi_key = str(row['InChI-Key'])
            inchi_key = inchi_key.removeprefix('InChIKey=') if 'InChIKey=' in inchi_key else inchi_key
            inchi_key = inchi_key if inchi_key != 'nan' else None  
            if inchi_key:
                add_cv_term_metabolites(inchi_key, 'InChI-Key', sp)
                
        if 'ChEBI' in missing_metabs_df.columns:
            chebi_id = str(int(row.get('ChEBI'))) if not math.isnan(float(row.get('ChEBI'))) else None 
            if chebi_id:
                add_cv_term_metabolites(chebi_id, 'ChEBI', sp)
                
    model = update_annotations_from_others(model)
    
    # (3) Add all missing reactions
    for _, row in missing_reacs_df.iterrows():
        reaction_dict = ast.literal_eval(str(row['bigg_reaction']))
        reactants = reaction_dict.get('reactants')
        products = reaction_dict.get('products')
        genes = ast.literal_eval(str(row['gene_product'])) if row['Spontaneous?'] != 'T' else 'G_spontaneous'
        compartment = row['compartment']
        compartment = compartment if compartment != 'exchange' else None
        reac, model = create_reaction(
            model=model_libsbml, reaction_id=row['bigg_id'], name=row['name'], reactants=reactants, 
            products=products, fluxes=ast.literal_eval(str(row['fluxes'])), compartment=compartment, genes=genes
            )
        if 'bigg_aliases' in missing_reacs_df.columns:
            bigg_aliases_row = ast.literal_eval(str(row['bigg_aliases']))
            if bigg_aliases_row:
                for bigg_id in bigg_aliases_row:
                    if bigg_id != reac.getId():
                        add_cv_term_reactions(bigg_id, 'BIGG', reac)
        
        if 'KEGG' in missing_reacs_df.columns:
            kegg_row = ast.literal_eval(str(row['KEGG']).replace('nan', 'None'))
            if kegg_row:
                for kegg_id in kegg_row:
                    if kegg_id:
                        add_cv_term_reactions(kegg_id, 'KEGG', reac)
                
        if 'BioCyc' in missing_reacs_df.columns:
            biocyc_row = ast.literal_eval(str(row['BioCyc']))
            if biocyc_row:
                for biocyc_id in biocyc_row:
                    add_cv_term_reactions(biocyc_id, 'BioCyc', reac)
                    add_cv_term_reactions(biocyc_id, 'METACYC', reac)
                    
        if 'EC' in missing_reacs_df.columns:
            ec_row = ast.literal_eval(str(row['EC']).replace('nan', 'None'))
            if ec_row:
                for ec_num in ec_row:
                    if ec_num:
                        add_cv_term_reactions(ec_num, 'EC', reac)
                         
    return model


def gapfill(
    model_libsbml: libModel, gapfill_params: dict[str: str], filename: str
    ) -> Union[tuple[pd.DataFrame, libModel], tuple[tuple, libModel]]:
    """| Main function to fill gaps in a model by comparing the locus tags of the GeneProducts to 
       | KEGG/BioCyc/(Genbank) GFF file

    Args:
        - model_libsbml (libModel): Model loaded with libSBML
        - gapfill_params (dict): Dictionary obtained from YAML file containing the parameter mappings
        - filename (str): Path to output file for gapfill analysis result
        - gapfill_model_out (str): Path where gapfilled model should be written to
        
    Returns:
        tuple: ``gap_analysis()`` table(s) (1) & libSBML model (2)
            (1) pd.DataFrame|tuple(pd.DataFrame): Result from function ``gap_analysis()``
            (2) libModel: Gap filled model
    """
    gap_analysis_result = gap_analysis(model_libsbml, gapfill_params, filename)
    model = gapfill_model(model_libsbml, gap_analysis_result)
    
    return gap_analysis_result, model
